#   Copyright (c) 2021 Boulder Real Time Technologies, Inc.
#
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#
#   Author: Ambli LLC (ambli.ll.company@gmail.com)

#master_stations 0.00000  9999999999.99900 css3.0          single AK  - /aec/db/stations master_stations 1624576260.73523
#catalog         55123200.00000  9999999999.99900 css3.0   month  -   - /aec/db/catalogs/final/monthly/%Y_%m   catalog_%Y_%m 1624576324.62700
#waveforms       852076800.00000  9999999999.99900 css3.0  day    -   - /aec/db/waveforms/%Y_%m   waveforms_%Y_%m_%d  -  1624576466.45122

import os
import sys
import glob
import logging
import datetime

sys.path.append(os.environ['ANTELOPE'] + "/data/python")
from antelope import stock
from antelope import datascope

from webservice_fdsn.app.core import constants
from webservice_fdsn.app.utils import utilities
from webservice_fdsn.app.utils import pf_helper
from webservice_fdsn.app.utils import exceptions

log = logging.getLogger(__name__)


# TODO: Make this an abc
class DBClusterRecord:
    def __init__(self, cluster_record):
        self.name = None
        self.schema = None
        self.network = None
        self.volumes = None
        self.directory = None
        self.dfile = None
        self.dfile_path = None
        self._epoch_start_time = None
        self._epoch_end_time = None
        self.start_time_dt = None
        self.end_time_dt = None

    def validate_record(self):
        raise NotImplementedException()

    def get_dfiles(self):
        raise NotImplementedException()

    def contains_requested_date_range(self):
        raise NotImplementedException()    


class DBCluster05(DBClusterRecord):
    """Class to handle objects conformant with dbcluster0.5 schema
    Implement if required
    """
    pass
	


class DBCluster07(DBClusterRecord):
    """Class to handle objects conformant with dbcluster0.7 schema
    """
    def __init__(self, cluster_record):
        """Open dbcentral db, store available clusters, and granularity, along with other metadata
        such as where the db files reside, schema etc
        """
        self.name = cluster_record.getv(constants.dbcluster_constants.CLUSTERNAME)[0]
        self.schema = cluster_record.getv(constants.dbcluster_constants.SCHEMA)[0]
        log.debug(f'Creating cluster object for {self.name} [Schema: {self.schema}]')
        self.network = cluster_record.getv(constants.dbcluster_constants.NET)[0]
        self.volumes = cluster_record.getv(constants.dbcluster_constants.VOLUMES)[0]
        self.directory = cluster_record.getv(constants.dbcluster_constants.DIR)[0]
        self.dfile = cluster_record.getv(constants.dbcluster_constants.DFILE)[0]
        self.dfile_path = os.path.join(self.directory, self.dfile)
        # This is the time from when records in this table start and end.
        self._epoch_start_time = cluster_record.getv(constants.dbcluster_constants.TIME)[0]
        self._epoch_end_time = cluster_record.getv(constants.dbcluster_constants.ENDTIME)[0]
        # Store datetime.datetime objects for easy consumption in module
        self.start_time_dt = datetime.datetime.fromtimestamp(self._epoch_start_time,
                                                             tz=datetime.timezone.utc)
        self.end_time_dt = datetime.datetime.fromtimestamp(self._epoch_end_time,
                                                           tz=datetime.timezone.utc)
        self.validate_record()

    def validate_record(self):
        """Any checks to be run on user supplied data
        """
        # Check whether absolute dir exists, remove wildcard segments which are dynamically
        # generated based on user request
        directory = self.directory
        while '%' in directory:
        	directory = os.path.dirname(directory)
        if not os.path.isdir(directory):
            raise Exception(f'Parent directory for cluster {self.name} does not exist: {directory}')

    def contains_requested_date_range(self, start_date, end_date):
        """Check whether user requested dates are available in this db.

        :param datetime start_date: window start time
        :param datetime end_date: window end time
        :rtype bool
        :return: true if db contains data matching user supplied time window
        """
        if end_date < self.start_time_dt or start_date > self.end_time_dt:
        	return False
        return True


    def get_dfiles(self, start_date: datetime.datetime, end_date: datetime.datetime,
    	           sanitize=False) -> list:
        """Return path to data files. If requested dates span multiple files, return all of them

        :param start_date: start date of window
        :param end_date: end date of window
        :param bool sanitize: If true, dont send back files which do not exist, else send everything in date range
        :rtype: list
        :return: list of paths of data files on disk. Could be an empty list is sanitize is true and no files were found
        """
        if self.volumes == constants.common_constants.SINGLE:
            return [stock.epoch2str(stock.str2epoch(format(start_date)), self.dfile_path)]

        dfiles = list()
        if self.volumes == constants.common_constants.DAY:
            # dfile will be of format <something>_%Y_%m_%d
            # If time window is less than 24 hours, but day/s roll/s over, need to include both
            # start and end days. 
            # Eg: If window is 2018/07/20T08:06:00->2018/07/21T01:00:00, (end-start).days will be 0.
            # But we really want to search both 2018/07/20 and 21st.
            # Create temp start and end days using original day/month/year but
            # drop the hour/minute/seconds
            window_start_date = datetime.datetime(year=start_date.year, month=start_date.month, day=start_date.day)
            window_end_date = datetime.datetime(year=end_date.year, month=end_date.month, day=end_date.day)
            num_days = (window_end_date - window_start_date).days

            # If end_date is same as start_date, return dfile for a single day
            if num_days == 0:
            	# stock.str2epoch will expand %-escaped sequences with actual date values
                dfiles.append(stock.epoch2str(stock.str2epoch(format(start_date)), self.dfile_path))
            else:
            	# num_days+1 so end is inclusive
                for x in range(num_days + 1):
                    df = stock.epoch2str(stock.str2epoch(format(start_date + datetime.timedelta(days=x))),
                                         self.dfile_path)
                    if df not in dfiles:
                        dfiles.append(df)

        elif self.volumes == constants.common_constants.MONTH:
            # dfile will be of format blah_%Y_%m
            selected = start_date
            # Keep adding files until the first out-of-range month/year is encountered.
            while (selected.month, selected.year) != ((end_date + datetime.timedelta(days=31)).month,
                                                      (end_date + datetime.timedelta(days=31)).year):
                df = stock.epoch2str(stock.str2epoch(f'{selected}'), self.dfile_path)
                if df not in dfiles:
                    dfiles.append(df)
				# Move ahead by 28 days which is the smallest month.
				# If moving ahead by 30 days, we might miss Feb in case where
				# selected is Jan 31st.. then +30 would be March 1st, thereby skipping Feb.
				# Using 28 will lead to duplicate entries, but that is fine because we are
				# aggregating using a set().
                selected += datetime.timedelta(days=28)

        elif self.volumes == constants.common_constants.YEAR:
            # dfile will be of format blah_%Y
            num_years = end_date.year - start_date.year
            # If end_year is same as start_date, return dfile for a single day
            if num_years == 0:
                dfiles.append(stock.epoch2str(stock.str2epoch(format(start_date)), self.dfile_path))
            else:
                for x in range(num_years):
                    df = stock.epoch2str(stock.str2epoch(format(start_date + datetime.timedelta(days=(x + 1) * 365))),
                                         self.dfile_path)
                    if df not in dfiles:
                        dfiles.append(df)

        else:
            raise Exception(f'Unsupported volume type of {self.volumes}')


        log.debug(f'List of generated dfiles: {",".join(dfiles)}')
        if sanitize:
            log.debug('Sanitization requested, eliminating non-existant dfile paths')
            sanitized = []
            for dfl in dfiles:
                if glob.glob(f'{dfl}*'):
                    sanitized.append(dfl)
                    log.debug(f'Found dfile: {dfl}')
            dfiles = sanitized

        return list(dfiles)


class DBClusterResolver:
	
    def __init__(self, dbcluster_path):
        """Helper class to assist with selecting correct data files by date from clusters on disk
        """
        if not os.path.exists(dbcluster_path):
            raise Exception(f'Invalid dbcluster path. "{dbcluster_path}" does not exist on disk or service does not \
                            have appropriate permissions to access it.')
        self.dbcluster_path = dbcluster_path

        _schema = utilities.get_db_schema(self.dbcluster_path)
        if _schema not in [constants.dbcluster_constants.DBCLUSTER07,
                           constants.dbcluster_constants.DBCLUSTER05]:
            raise Exception(f'Unsupported dbcluster schema {_schema}')
        else:
            self.schema = _schema

        log.debug(f'Initializing object with {dbcluster_path}, schema: {self.schema}')
        dbcluster = datascope.dbopen(dbcluster_path, 'r', 1)
        try:
            clusters = dbcluster.lookup(table='clusters')
        except antelope.datascope.DblookupFieldError:
            raise Exception(f'Table "clusters" not found in db')
        log.debug(f'DB {dbcluster_path} has {clusters.record_count} records')

        # Create metadata objects for each cluster record
        self.clusters = {}
        SchemaClass = DBCluster07 if self.schema == constants.dbcluster_constants.DBCLUSTER07 else DBCluster05
        for r in clusters.iter_record():
            _clstr = SchemaClass(r)
            self.clusters[_clstr.name] = _clstr
        log.debug(f'Clusters in dbcentral: {self.clusters.keys()}')
	
    def get_dfiles_by_date(self, fdsn_api_group, start_date, end_date=None,
    	                   sanitize=False):
        """Return path to data file based on supplied dates

        :param str fdsn_api_group: name of cluster to look for
        :param datetime.datetime start_date:
        :param datetime.datetime end_date:
        :param bool sanitize: If true, dont send back files which do not exist, else send everything in date range
        :rtype: str
        :return: path to data file
        """
        # Users must supply name of cluster they have assigned to dbcluster in the pf file
        # Else we have no idea which entry corresponds to which api group.
        cluster_name = pf_helper.get_cluster_name_for_api(fdsn_api_group)

        if not end_date:
            end_date = datetime.datetime.now()

        if end_date < start_date:
            raise Exception('End date cannot be before start date')

        log.debug(f'Looking for dfiles in time window [{start_date} -> {end_date}]')
        return self.clusters[cluster_name].get_dfiles(start_date, end_date, sanitize=sanitize)
